// Precompiled headers for VC++
#include "stdafx.h"
#include "Brain.h"
#include <algorithm>
#include <functional>
#include <vector>
#include <cassert>

LobeData::LobeData() 
{
	fCell = 0;
	std::fill(fData, fData + kLast, 0);
}

LobeData::LobeData(unsigned int cell, 
	unsigned int output, 
	unsigned int state,
	unsigned int dendrites,
	unsigned int susceptability,
	unsigned int stw,
	unsigned int ltw,
	unsigned int strength)
{
	fCell = cell;
	fData[kOutput] = output;
	fData[kState] = state;
	fData[kNumberOfDendrites] = dendrites;
	fData[kTotalSusceptability] = susceptability;
	fData[kTotalSTW] = stw;
	fData[kTotalLTW] = ltw;
	fData[kTotalStrength] = strength;
}

int outputIsZero(LobeData rhs)
{
	return rhs.getOutput() == 0;
}

int stateIsZero(LobeData rhs)
{
	return rhs.getState() == 0;
}

int outputIsGreater(LobeData const& lhs, LobeData const& rhs)
{
	return lhs.getOutput() > rhs.getOutput();
}

int stateIsGreater(LobeData const& lhs, LobeData const& rhs)
{
	return lhs.getState() > rhs.getState();
}


std::string Lobe::getMacroCommand() const
{
	char buffer[128];
	wsprintf(buffer, 
		"inst,setv obv1 0 loop dde: cell %d obv1 0 addv obv1 1 untl obv1 eq %d,endm",
		getLobeNumber(),
		getNumberOfCells());
	return buffer;
}

Lobe::CellVector Lobe::parseCellData(std::string const& result) const
{
	CellVector cells;
	int position = 0;
	const int numberOfCells = getNumberOfCells();
	for(int i = 0; i < numberOfCells; ++i)
	{
		LobeData data;
		data.fCell = i;
		for(int j = 0; j < LobeData::kLast; ++j)
		{
			data.fData[j] = atoi(result.substr(position).c_str());
			position = result.find("|", position) + 1;
		}
		cells.push_back(data);
	}
	return cells;
}

int SensoryLobe::getLobeNumber() const
{
	return 5;
}

std::string SensoryLobe::getCellDescription(int cell) const
{
	static char const* cellNames[] =
	{
		"I've been patted",
		"I've been slapped",
		"I've bumped into a wall",
		"I am near a wall",
		"I am in a vehicle",
		"User has spoken",
		"Creature has spoken",
		"Own kind has spoken",
		"Audible event",
		"Visible event",
		"It is approaching",
		"It is retreating",
		"It is near me",
		"It is active",
		"It is an object",
		"It is a creature",
		"It is my sibling",
		"It is my parent",
		"It is my child",
		"It is opposite sex",
		"spare2",
		"spare3",
		"spare4",
		"spare5",
		"spare6",
		"spare7",
		"spare8",
		"spare9",
		"spare10",
		"spare11",
		"spare12",
		"spare13"
	};

	return cellNames[cell];
}

unsigned int SensoryLobe::getNumberOfCells() const
{
	return 32;
}

int PerceptionLobe::getLobeNumber() const
{
	return 0;
}

std::string PerceptionLobe::getCellDescription(int cell) const
{

	// Modified to return the cell description based upon
	// the lobe mappings in a standard norn.	
	static char const* cellNames[] =
	{
		"Pain",
		"Need for pleasure",
		"Hunger",
		"Coldness",
		"Hotness",
		"Tiredness",
		"Sleepiness",
		"Loneliness",
		"Overcrowdedness",
		"Fear",
		"Boredom",
		"Anger",
		"Sex drive",
		"Drive13",
		"Drive14",
		"Drive15",
		"Quiescent",
		"Activate 1 it (push)",
		"Activate 2 it (pull)",
		"Deactivate it (stop)",
		"Approach it (come)",
		"Retreat from it (run)",
		"Get it (get)",
		"Drop all (drop)",
		"Say what you need (think)",
		"Rest (sleep)",
		"Travel west (left)",
		"Travel east (right)",
		"Verb12",
		"Verb13",
		"Verb14",
		"Verb15",
		"I've been patted",
		"I've been slapped",
		"I've bumped into a wall",
		"I am near a wall",
		"I am in a vehicle",
		"User has spoken",
		"Creature has spoken",
		"Own kind has spoken",
		"Audible event",
		"Visible event",
		"It is approaching",
		"It is retreating",
		"It is near me",
		"It is active",
		"It is an object",
		"It is a creature",
		"It is my sibling",
		"It is my parent",
		"It is my child",
		"It is opposite sex",
		"sense20",
		"sense21",
		"sense22",
		"sense23",
		"sense24",
		"sense25",
		"sense26",
		"sense27",
		"sense28",
		"sense29",
		"sense30",
		"sense31",
		"Myself (norn)",
		"Hand",
		"Call",
		"Water",
		"Plant",
		"Egg",
		"Food",
		"Drink",
		"Vendor",
		"Music",
		"Animal",
		"Fire",
		"Shower",
		"Toy",
		"Bigtoy",
		"Weed",
		"Word32",
		"Word33",
		"Word34",
		"Word35",
		"Word36",
		"Word37",
		"Word38",
		"Word39",
		"Word40",
		"Word41",
		"Mover",
		"Lift",
		"Computer",
		"Fun",
		"Bang",
		"Word47",
		"Word48",
		"Word49",
		"Word50",
		"Word51",
		"Norn",
		"Grendle",
		"Word54",
		"Word55"
	};

	static const int size = sizeof(cellNames) / sizeof(cellNames[0]);
	char buffer[30];
	std::string retval = (cell >= size) ? (std::string("Cell number ") + itoa(cell, buffer, 10)) : cellNames[cell];
	return retval;
}

unsigned int PerceptionLobe::getNumberOfCells() const
{
	return 112;
}

int ConceptLobe::getLobeNumber() const
{
	return 8;
}

std::string ConceptLobe::getCellDescription(int cell) const
{
	char buffer[30];
	return std::string("Cell number ") + itoa(cell, buffer, 10);
}

unsigned int ConceptLobe::getNumberOfCells() const
{
	return 640;
}

int DecisionLobe::getLobeNumber() const
{
	return 6;
}

std::string DecisionLobe::getCellDescription(int cell) const
{
	static char const* cellNames[] =
	{
		"Quiescent",
		"Activate 1 it (push)",
		"Activate 2 it (pull)",
		"Deactivate it (stop)",
		"Approach it (come)",
		"Retreat from it (run)",
		"Get it (get)",
		"Drop all (drop)",
		"Say what you need (think)",
		"Rest (sleep)",
		"Travel west (left)",
		"Travel east (right)",
		"Verb 12",
		"Verb 13",
		"Verb 14",
		"Verb 15"
	};

	return cellNames[cell];
}

unsigned int DecisionLobe::getNumberOfCells() const
{
	return 16;
}

int AttentionLobe::getLobeNumber() const
{
	return 7;
}

std::string AttentionLobe::getCellDescription(int cell) const
{
	static char const* cellNames[] =
	{
		"Myself (norn)",
		"Hand",
		"Call",
		"Water",
		"Plant",
		"Egg",
		"Food",
		"Drink",
		"Vendor",
		"Music",
		"Animal",
		"Fire",
		"Shower",
		"Toy",
		"Bigtoy",
		"Weed",
		"Word32",
		"Word33",
		"Word34",
		"Word35",
		"Word36",
		"Word37",
		"Word38",
		"Word39",
		"Word40",
		"Word41",
		"Mover",
		"Lift",
		"Computer",
		"Fun",
		"Bang",
		"Word47",
		"Word48",
		"Word49",
		"Word50",
		"Word51",
		"Norn",
		"Grendle",
		"Word54",
		"Word55"
	};

	return cellNames[cell];
}

unsigned int AttentionLobe::getNumberOfCells() const
{
	return 40;
}

int NounLobe::getLobeNumber() const
{
	return 4;
}

int VerbLobe::getLobeNumber() const
{
	return 3;
}

int StimulusSourceLobe::getLobeNumber() const
{
	return 2;
}

#if 0
std::string StimulusSourceLobe::getCellDescription(int cell) const
{
	static char const* cellNames[] =
	{
		"Disappointment",
		"Pointer pats me",
		"Creature pats me",
		"Pointer slaps me",
		"Creature slaps me",
		"It is approaching",
		"It is retreating",
		"I bump into a wall",
		"Object comes into view",
		"Unrecognised word",
		"Heard user speak",
		"Heard creature speak",
		"I am quiescient (periodic)",
		"I've activated1",
		"I've activated2",
		"I've deactivated",
		"I am approaching (periodic)",
		"I have retreated",
		"I have got",
		"I have dropped",
		"I've stated need",
		"I am resting (periodic)",
		"I am sleeping (periodic)",
		"I am travelling (periodic)",
		"Spare1",
		"Spare2",
		"Spare3",
		"Spare4",
		"Involuntary0",
		"Involuntary1 (laying)",
		"Involuntary2 (sneezing)",
		"Involuntary3 (coughing)",
		"Involuntary4 (shivering)",
		"Involuntary5",
		"Involuntary6",
		"Involuntary7",
		"Unknown1",
		"Unknown2",
		"Unknown3",
		"Unknown4"
	};

	return cellNames[cell];
}
#endif
unsigned int StimulusSourceLobe::getNumberOfCells() const
{
	return 40;
}

int DriveLobe::getLobeNumber() const
{
	return 1;
}

std::string DriveLobe::getCellDescription(int cell) const
{
	static char const* cellNames[] =
	{
		"Pain",
		"Need for pleasure",
		"Hunger",
		"Coldness",
		"Hotness",
		"Tiredness",
		"Sleepiness",
		"Loneliness",
		"Overcrowdedness",
		"Fear",
		"Boredom",
		"Anger",
		"Sex drive",
		"Not allocated2",
		"Not allocated3",
		"Not allocated4"
	};

	return cellNames[cell];
}

unsigned int DriveLobe::getNumberOfCells() const
{
	return 16;
}

std::string CompositeLobe::getMacroCommand() const 
{ 
	std::string header = "inst,";
	std::string footer = ",endm";
	std::string command = header;
	char buffer[30];

	ItemList::const_iterator it = fItems.begin();
	while(it != fItems.end())
	{
		Item const& item = (*it);
		wsprintf(buffer, "dde: cell %d %d %d", 
			item.fLobe,
			item.fCell,
			item.fDendrite);
		command += buffer;
		if(++it != fItems.end())
		{
			command += " ";
		}
	}
	command += footer;
	return command;
}

std::string CompositeLobe::getCellDescription(int cell) const
{
	return fItems[cell].fDescription;
}

unsigned int CompositeLobe::getNumberOfCells() const
{
	return fItems.size();
}

void CompositeLobe::addCell(unsigned int lobe, unsigned int cell, unsigned int dendrite, std::string const& desc)
{
	fItems.push_back(Item(lobe, cell, dendrite, desc));
}

void CompositeLobe::removeCell(unsigned int index)
{
	fItems.erase(fItems.begin() + index);
}

Brain::Brain()
{
	PerceptionLobe perception;
	DriveLobe drive;
	StimulusSourceLobe source;
	VerbLobe verb;
	NounLobe noun;
	SensoryLobe sense;
	DecisionLobe decision;
	AttentionLobe attention;
	ConceptLobe concept;

	int w = 0;
	int h = 0;
	int cell = 0;
	std::string name = "Perception";

	// Perception
	for(h = 0; h < 16; ++h)
	{
		for(w = 0; w < 7; ++w)
		{
			fCells[4 + w][13 + h] = BrainCell(name, 
				perception.getCellDescription(cell++));
		}
	}

	// Drive
	w = 0; h = 0; cell = 0; name = "Drive";
	for(h = 0; h < 2; ++h)
	{
		for(w = 0; w < 8; ++w)
		{
			fCells[34 + w][30 + h] = BrainCell(name, 
				drive.getCellDescription(cell++));
		}
	}

	// Source
	w = 0; h = 0; cell = 0; name = "Source";
	for(h = 0; h < 5; ++h)
	{
		for(w = 0; w < 8; ++w)
		{
			fCells[15 + w][24 + h] = BrainCell(name, 
				source.getCellDescription(cell++));
		}
	}

	// Verb
	w = 0; h = 0; cell = 0; name = "Verb";
	for(h = 0; h < 2; ++h)
	{
		for(w = 0; w < 8; ++w)
		{
			fCells[37 + w][24 + h] = BrainCell(name, 
				verb.getCellDescription(cell++));
		}
	}

	// Noun
	w = 0; h = 0; cell = 0; name = "Noun";
	for(h = 0; h < 2; ++h)
	{
		for(w = 0; w < 20; ++w)
		{
			fCells[21 + w][3 + h] = BrainCell(name, 
				noun.getCellDescription(cell++));
		}
	}

	// GeneralSensory
	w = 0; h = 0; cell = 0; name = "General Sense";
	for(h = 0; h < 4; ++h)
	{
		for(w = 0; w < 8; ++w)
		{
			fCells[32 + w][34 + h] = BrainCell(name, 
				sense.getCellDescription(cell++));
		}
	}

	// Decision
	w = 0; h = 0; cell = 0; name = "Decision";
	for(h = 0; h < 16; ++h)
	{
		for(w = 0; w < 1; ++w)
		{
			fCells[53 + w][15 + h] = BrainCell(name, 
				decision.getCellDescription(cell++));
		}
	}

	// Attention
	w = 0; h = 0; cell = 0; name = "Attention";
	for(h = 0; h < 8; ++h)
	{
		for(w = 0; w < 5; ++w)
		{
			fCells[44 + w][30 + h] = BrainCell(name, 
				attention.getCellDescription(cell++));
		}
	}

	// Concept
	w = 0; h = 0; cell = 0; name = "Concept";
	for(h = 0; h < 16; ++h)
	{
		for(w = 0; w < 40; ++w)
		{
			fCells[12 + w][6 + h] = BrainCell(name, 
				concept.getCellDescription(cell++));
		}
	}
}

std::string Brain::getLobeName(unsigned int x, unsigned int y)
{
	return fCells[x][y].fLobeName;
}

std::string Brain::getCellDescription(unsigned int x, unsigned int y)
{
	return fCells[x][y].fCellName;
}

Brain::ResultsVector Brain::parseFromActivity(std::string const& activity)
{
	ResultsVector results;
	unsigned int index = 0;
	while(index < activity.length())
	{
		unsigned int x = activity[index++] - '0';
		assert(index < activity.length());
		unsigned int y = activity[index++] - '0';
		assert(index < activity.length());
		unsigned int value = activity[index++] - '0';
		results.push_back(Results(x, y, value));
	}
	return results;
}
